@@ -1,23 +1,30 @@
-""" Defines a KernelClient that provides thread-safe sockets with async callbacks on message replies.
+""" Defines a KernelClient that provides thread-safe sockets with async callbacks on message
+replies.
 """
-from __future__ import absolute_import
 import atexit
 import errno
-from threading import Thread
+import sys
 import time
+from threading import Event
+from threading import Thread
 
-# import ZMQError in top-level namespace, to avoid ugly attribute-error messages
-# during garbage collection of threads at exit:
+from traitlets import Instance
+from traitlets import Type
 from zmq import ZMQError
-from zmq.eventloop import ioloop, zmqstream
+from zmq.eventloop import ioloop
+from zmq.eventloop import zmqstream
+
+from jupyter_client import KernelClient
+from jupyter_client.channels import HBChannel
 
 # Local imports
-from traitlets import Type, Instance
-from jupyter_client.channels import HBChannel
-from jupyter_client import KernelClient
+# import ZMQError in top-level namespace, to avoid ugly attribute-error messages
+# during garbage collection of threads at exit
+
 
 class ThreadedZMQSocketChannel(object):
     """A ZMQ socket invoking a callback in the ioloop"""
+
     session = None
     socket = None
     ioloop = None
@@ -36,16 +43,23 @@ class ThreadedZMQSocketChannel(object):
         loop
             A pyzmq ioloop to connect the socket to using a ZMQStream
         """
-        super(ThreadedZMQSocketChannel, self).__init__()
+        super().__init__()
 
         self.socket = socket
         self.session = session
         self.ioloop = loop
+        evt = Event()
 
-        self.stream = zmqstream.ZMQStream(self.socket, self.ioloop)
-        self.stream.on_recv(self._handle_recv)
+        def setup_stream():
+            self.stream = zmqstream.ZMQStream(self.socket, self.ioloop)
+            self.stream.on_recv(self._handle_recv)
+            evt.set()
+
+        self.ioloop.add_callback(setup_stream)
+        evt.wait()
 
     _is_alive = False
+
     def is_alive(self):
         return self._is_alive
 
@@ -73,8 +87,10 @@ class ThreadedZMQSocketChannel(object):
         This is threadsafe, as it uses IOLoop.add_callback to give the loop's
         thread control of the action.
         """
+
         def thread_send():
             self.session.send(self.stream, msg)
+
         self.ioloop.add_callback(thread_send)
 
     def _handle_recv(self, msg):
@@ -82,7 +98,7 @@ class ThreadedZMQSocketChannel(object):
 
         Unpacks message, and calls handlers with it.
         """
-        ident,smsg = self.session.feed_identities(msg)
+        ident, smsg = self.session.feed_identities(msg)
         msg = self.session.deserialize(smsg)
         # let client inspect messages
         if self._inspect:
@@ -105,7 +121,6 @@ class ThreadedZMQSocketChannel(object):
         """
         pass
 
-
     def flush(self, timeout=1.0):
         """Immediately processes all pending messages on this channel.
 
@@ -139,22 +154,44 @@ class ThreadedZMQSocketChannel(object):
 
 
 class IOLoopThread(Thread):
-    """Run a pyzmq ioloop in a thread to send and receive messages
-    """
-    _exiting = False
+    """Run a pyzmq ioloop in a thread to send and receive messages"""
 
-    def __init__(self, loop):
-        super(IOLoopThread, self).__init__()
+    _exiting = False
+    ioloop = None
+
+    def __init__(self):
+        super().__init__()
         self.daemon = True
-        self.ioloop = loop or ioloop.IOLoop()
 
     @staticmethod
     @atexit.register
     def _notice_exit():
-        IOLoopThread._exiting = True
+        # Class definitions can be torn down during interpreter shutdown.
+        # We only need to set _exiting flag if this hasn't happened.
+        if IOLoopThread is not None:
+            IOLoopThread._exiting = True
+
+    def start(self):
+        """Start the IOLoop thread
+
+        Don't return until self.ioloop is defined,
+        which is created in the thread
+        """
+        self._start_event = Event()
+        Thread.start(self)
+        self._start_event.wait()
 
     def run(self):
         """Run my loop, ignoring EINTR events in the poller"""
+        if "asyncio" in sys.modules:
+            # tornado may be using asyncio,
+            # ensure an eventloop exists for this thread
+            import asyncio
+
+            asyncio.set_event_loop(asyncio.new_event_loop())
+        self.ioloop = ioloop.IOLoop()
+        # signal that self.ioloop is defined
+        self._start_event.set()
         while True:
             try:
                 self.ioloop.start()
@@ -179,9 +216,10 @@ class IOLoopThread(Thread):
         :meth:`~threading.Thread.start` is called again.
         """
         if self.ioloop is not None:
-            self.ioloop.stop()
+            self.ioloop.add_callback(self.ioloop.stop)
         self.join()
         self.close()
+        self.ioloop = None
 
     def close(self):
         if self.ioloop is not None:
@@ -192,36 +230,31 @@ class IOLoopThread(Thread):
 
 
 class ThreadedKernelClient(KernelClient):
-    """ A KernelClient that provides thread-safe sockets with async callbacks on message replies.
-    """
+    """A KernelClient that provides thread-safe sockets with async callbacks on message replies."""
 
-    _ioloop = None
     @property
     def ioloop(self):
-        if self._ioloop is None:
-            self._ioloop = ioloop.IOLoop()
-        return self._ioloop
+        return self.ioloop_thread.ioloop
 
     ioloop_thread = Instance(IOLoopThread, allow_none=True)
 
-    def start_channels(self, shell=True, iopub=True, stdin=True, hb=True):
+    def start_channels(self, shell=True, iopub=True, stdin=True, hb=True, control=True):
+        self.ioloop_thread = IOLoopThread()
+        self.ioloop_thread.start()
+
         if shell:
             self.shell_channel._inspect = self._check_kernel_info_reply
 
-        self.ioloop_thread = IOLoopThread(self.ioloop)
-        self.ioloop_thread.start()
-
-        super(ThreadedKernelClient, self).start_channels(shell, iopub, stdin, hb)
+        super().start_channels(shell, iopub, stdin, hb, control)
 
     def _check_kernel_info_reply(self, msg):
-        """This is run in the ioloop thread when the kernel info reply is received
-        """
-        if msg['msg_type'] == 'kernel_info_reply':
+        """This is run in the ioloop thread when the kernel info reply is received"""
+        if msg["msg_type"] == "kernel_info_reply":
             self._handle_kernel_info_reply(msg)
             self.shell_channel._inspect = None
 
     def stop_channels(self):
-        super(ThreadedKernelClient, self).stop_channels()
+        super().stop_channels()
         if self.ioloop_thread.is_alive():
             self.ioloop_thread.stop()
 
@@ -229,3 +262,4 @@ class ThreadedKernelClient(KernelClient):
     shell_channel_class = Type(ThreadedZMQSocketChannel)
     stdin_channel_class = Type(ThreadedZMQSocketChannel)
     hb_channel_class = Type(HBChannel)
+    control_channel_class = Type(ThreadedZMQSocketChannel)
